/*----------------------------------------------------------------------------
 Copyright:      
 Author:         Radig Ulrich
 Remarks:        
 known Problems: none
 Version:        20.01.2013
 Description:  

 Dieses Programm ist freie Software. Sie knnen es unter den Bedingungen der 
 GNU General Public License, wie von der Free Software Foundation verffentlicht, 
 weitergeben und/oder modifizieren, entweder gem Version 2 der Lizenz oder 
 (nach Ihrer Option) jeder spteren Version. 

 Die Verffentlichung dieses Programms erfolgt in der Hoffnung, 
 da es Ihnen von Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE, 
 sogar ohne die implizite Garantie der MARKTREIFE oder der VERWENDBARKEIT 
 FR EINEN BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License. 

 Sie sollten eine Kopie der GNU General Public License zusammen mit diesem 
 Programm erhalten haben. 
 Falls nicht, schreiben Sie an die Free Software Foundation, 
 Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.   
------------------------------------------------------------------------------*/
#define F_CPU 12000000
#define DMX_BAUD 250000
#define DMX_LOST_TIMEOUT 500
#define DMX_BAUD_BREAK 80000

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>
#include "rf22.h"

volatile unsigned char dmx_buffer[514];
volatile unsigned char dmx_tmp_buffer[33];
volatile unsigned char new_dmx_data = 0;

volatile unsigned int dmx_lost = DMX_LOST_TIMEOUT;
 

#define TRANSMIT_BLOCKS 16	//1 Block 32Channels max.16 Blocks 512Channels
#define TRANSMIT_REPEAT	64

//############################################################################
//DMX Senderoutine
ISR (USART_TX_vect)
//############################################################################
{
	static unsigned int  dmx_channel_tx_count = 1;
	static unsigned char dmx_tx_state = 0;
	
	switch (dmx_tx_state)
	{
		case (0):
			UBRR0   = (F_CPU / (DMX_BAUD_BREAK * 16L) - 1);
			UDR0 = 0; //RESET Frame
			dmx_tx_state = 1;
			break;

		case (1):
			UBRR0   = (F_CPU / (DMX_BAUD * 16L) - 1);
			UDR0 = 0; //Start Byte
			dmx_tx_state = 2;
			break;

		case (2):
			_delay_us(10);
			//Ausgabe des Zeichens
			UDR0 = dmx_buffer[dmx_channel_tx_count];
			dmx_channel_tx_count++;
			
			if(dmx_channel_tx_count == 512)
			{
				dmx_channel_tx_count = 1;
				dmx_tx_state = 0;
			}
			break;
	}
}


//############################################################################
//DMX Empfangsroutine (RS485)
ISR (USART_RX_vect)
//############################################################################
{
	static unsigned int dmx_channel_rx_count = 0;
	static unsigned char dmx_valid = 0;
	unsigned char tmp = 0;
	
	tmp =  UDR0;
	
	if(UCSR0A&(1<<FE0))
	{
		if(dmx_channel_rx_count > 1)
		{
			dmx_lost = 0;
		}
		dmx_channel_rx_count = 0;	
		dmx_buffer[0] = tmp;
		
		if(dmx_buffer[0] == 0)
		{
			dmx_valid = 1;
			dmx_channel_rx_count++;
		}
		else
		{
			dmx_valid = 0;
		}
		return;
	}
	
	if(dmx_valid)
	{
		if(dmx_buffer[dmx_channel_rx_count] != tmp)
		{
			dmx_buffer[dmx_channel_rx_count] = tmp;	
			new_dmx_data = TRANSMIT_REPEAT;
		}
	
		if(dmx_channel_rx_count < 514)
		{
			dmx_channel_rx_count++;
		}
		return;
	}
}

//############################################################################
//Hier wird die Zeit gezhlt (Tick 1ms)
ISR (TIMER2_COMPA_vect)
//############################################################################
{
	if(dmx_lost<DMX_LOST_TIMEOUT)
	{
		dmx_lost++;
	}
}

//############################################################################
//Hauptprogramm DMX FUNK BERTRAGUNG
int main(void)
//############################################################################
{	
	unsigned int tmp;
	unsigned char packet_cnt = 0;
	unsigned long counter = 0;
	DDRB |= (1<<PB0); //LED PORT Output
	DDRD |= (1<<PD3); //Direction RS485 RX/TX
	
	PORTB |= (1<<0); //LED ON
	
	PORTC |= (1<<PC0);//Pullup (Receiver/Transmitter)
	for(unsigned long b = 0;b<100000;b++){;asm("nop");};
	
	rf22_init();
	rf22_setfreq(RF22FREQ(869.5));
	rf22_rxmode();	
	
	
	if(PINC & (1<<PC0)) //Module als Sender oder Empfnger
	{
		PORTD &= ~(1<<PD3); //TX RS485
	
		//Init usart DMX-BUS
		UBRR0   = (F_CPU / (DMX_BAUD * 16L) - 1);
		UCSR0B|=(1 << RXEN0 | 1<< RXCIE0);
		UCSR0C|=(1<<USBS0); //USBS0 2 Stop bits	
		sei();//Globale Interrupts Enable
	
		//Timer2 Timecounter fr DMX Ausfall Kabel
		TCCR2A |= (1<<WGM21);
		TCCR2B |= (1<<CS22|1<<CS21|1<<CS20);
		TIMSK2 |= (1<<OCIE2A);
		OCR2A = F_CPU/1024/1000 - 1; //Tick Time
	
		while(1)
		{
			if(new_dmx_data > 0)
			{
				new_dmx_data--;	
				//512 Kanle werden zu je 32Bytes a 16 Packete versendet
				memcpy((unsigned char*)&dmx_tmp_buffer[1],(unsigned char *)&dmx_buffer[(packet_cnt*32)],32);
				dmx_tmp_buffer[0] = packet_cnt;
				packet_cnt++;
				
				if(packet_cnt == TRANSMIT_BLOCKS)
				{
					packet_cnt = 0;
				}
				
				rf22_sendpacket(dmx_tmp_buffer,33); //Packetcounter und 32DMX Bytes
			}
			
			//Abfrage ob DMX Packete via Kabel empfangen wurden
			if(dmx_lost==DMX_LOST_TIMEOUT)
			{
				PORTB &=~(1<<0); //LED AUS NO DMX
			}
			else
			{
				PORTB |= (1<<0); //LED AN DMX OK
			}	
		}
	}
	else
	{
		PORTD |= (1<<PD3); //RX RS485
		
		//Timer2 Timecounter fr DMX Ausfall Funk
		TCCR2A |= (1<<WGM21);
		TCCR2B |= (1<<CS22|1<<CS21|1<<CS20);
		TIMSK2 |= (1<<OCIE2A);	
		OCR2A = F_CPU/1024/1000 - 1; //Tick Time
		
		//Init usart DMX-BUS
		UBRR0   = (F_CPU / (DMX_BAUD * 16L) - 1);
		DDRD |= (1<<PD1); //Output TXD Pin ATmega88
		UCSR0B|=(1<<TXEN0)|(1<<TXCIE0); // TXEN0 Transmitter enable / TXCIE0 TX complete interrupt enable 
		UCSR0C|=(1<<USBS0); //USBS0 2 Stop bits	
		sei();//Globale Interrupts Enable
		UDR0 = 0;//Start DMX
		
		while(1)
		{	
			//Warte bis Packet empfangen wurde
			while(PIND&(1<<2))
			{
				if(dmx_lost == DMX_LOST_TIMEOUT)
				{
					PORTB &= ~(1<<0); //Status LED aus
					counter++;
					if(counter >1000000)
					{
						PORTB |= (1<<0);
						for(unsigned a = 0;a<5000;a++){;asm("nop");};
						PORTB &= ~(1<<0);
						counter = 0;
					}
				}
				else
				{
					PORTB |= (1<<0); //Status LED an
					counter = 0;
				}
			}

			tmp = rf22_getpacket(dmx_tmp_buffer); //Packetcounter und 32DMX Bytes
		
			//wurden alle Bytes empfangen?
			if(tmp == 33)
			{
				dmx_lost= 0;
				//512 Kanle werden zu je 32Bytes a 16 Packete empfangen
				packet_cnt = dmx_tmp_buffer[0];
				memcpy((unsigned char*)&dmx_buffer[(packet_cnt*32)],(unsigned char *)&dmx_tmp_buffer[1],32);
			}
		}
	}
}
